/* 
 *
 * With the power of GPL, Claudio Agosti <vecna@s0ftpj.org> proud to 
 * cut & paste code of:
 *
 * 2000-2001 by Harald Welte <laforge@gnumonks.org>
 *
 * develop SNIFFJOKE project using ulogd, ulogd is a COOOL framework able to 
 * steal packets from the kernel, and so manipulate without travel on 
 * dangerous kernel space memory and panic-like function & mad skbuff,
 *
 * I'm pleased to use so nice tool, when years ago I've approach an evil
 * solution with Innova project ( http://www.s0ftpj.org/project/innova )
 * and ulogd give me a valid alternative to avoid that scramblous hack.
 *
 * as heritage, this code follow GPL 2 license:
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 
 *  as published by the Free Software Foundation
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  -------------------------------------------------------------------
 *
 *  documentation is provided on packet distribution.
 *  this is a TEST SOFTWARE, in a TEST TECHNOLOGY, in a TEST IMPLEMENTATION
 *
 *  so, don't loose your time searching for UML code :)
 *  you should find information in http://www.s0ftpj.org and 
 *  http://www.delirandom.net/category/EN, or, if you are italian
 *  http://www.delirandom.net/category/IT
 *
 * This project link also some code from BLASTERSTEG,
 * http://www.s0ftpj.org/tools/blastersteg.tar.gz, but I don't give credit to 
 * unknow.
 *
 * Thanks to my roommate hyperion for him tech-nerdish blood.
 *
 * Thanks to manowar and lorenzo/jovanotti too
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <time.h>
#include <pwd.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netinet/ip_icmp.h>
#include <ulogd/ulogd.h>
#include <ulogd/conffile.h>

extern int errno;
static FILE *log;

/* 
 * number of hack provided from SNIFFJOKE module,
 * this value is required for understand when
 * all hacks are used jet
 */
#define SNIFFJOKE_HACK_NUMBER	8

static void sniffjokelog(const char *format, ...)
{
	char buf[1024];
	static int cnt;
	va_list ap;

	va_start(ap, format);

	if(log != NULL)
	{
		fprintf(log, "%d ", cnt++);
		vfprintf(log, format, ap);
	}
	else
	{
		vsnprintf(buf, 1024, format, ap);
		ulogd_log(ULOGD_NOTICE, "sniffjoke notice [%d %s]", cnt++, buf);
	}
	va_end(ap);

	fflush(log);
}

#define SNIFFJOKE_DEFAULT_LOG 	"/tmp/sniffjoke.log"
static config_entry_t fileopt_ce = { 
        .key = "file",
	.type = CONFIG_TYPE_STRING,
	.options = CONFIG_OPT_NONE,
	.u = { .string = SNIFFJOKE_DEFAULT_LOG }
};

#define SNIFFJOKE_DEFAULT_PORT	"1:65535"
static config_entry_t portopt_ce = {
	.key = "port",
	.type = CONFIG_TYPE_STRING,
	.options = CONFIG_OPT_NONE,
	.u = { .string = SNIFFJOKE_DEFAULT_PORT }
};

#define SNIFFJOKE_DEFAULT_USER	"invalid_default_sniffjoke"
static config_entry_t useropt_ce = {
	.key = "user",
	.type = CONFIG_TYPE_STRING,
	.options = CONFIG_OPT_NONE,
	.u = { .string = SNIFFJOKE_DEFAULT_USER }
};

#define CMDLINE	4096
static char **ipt_put_tcp =NULL, ipt_put_icmp[CMDLINE], 
     	    **ipt_remove_tcp =NULL, ipt_remove_icmp[CMDLINE]
     	;
static int rules;

struct intr_id {
	char *name;
	unsigned int id;
};

static int sockfd;

#define INTR_IDS        6
static struct intr_id intr_ids[INTR_IDS] = 
{
        { "raw.pkt", 0 },
	{ "raw.pktlen", 0 },
	{ "ip.totlen", 0 },
	{ "oob.time.sec", 0 },
	{ "oob.time.usec", 0 },
	{ "oob.mark", 0 },
};

#define GET_VALUE(x)    ulogd_keyh[intr_ids[x].id].interp->result[ulogd_keyh[intr_ids[x].id].offset].value
#define GET_FLAGS(x)    ulogd_keyh[intr_ids[x].id].interp->result[ulogd_keyh[intr_ids[x].id].offset].flags

inline unsigned int half_cksum(void *pointed_data, int len)
{
	unsigned int sum = 0x00;
	unsigned short carry = 0x00;
	/*
	 * this strange (any pointer *) -> (void *) -> (unsigned short *) 
	 * is an hack to avoid:
	 *
	 *
warning: dereferencing type-punned pointer will break strict-aliasing rules

   ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ type-punned !?!?	 
	 */
	unsigned short *data =(unsigned short *)pointed_data;

	while (len > 1)
	{
		sum += *data++;
		len -= 2;
	}

	if (len == 1)
	{
		*((unsigned short *) &carry) = *(unsigned char *) data;
		sum += carry;
	}

	return sum;
}

inline unsigned short compute_sum(unsigned int sum)
{
	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);

	return (unsigned short) ~sum;
}

static inline int truncate_tcp_data(struct tcphdr *tcph, struct iphdr *iph)
{
	tcph->doff =sizeof(struct tcphdr) / 4; 
	iph->tot_len =htons(iph->ihl * 4 + sizeof(struct tcphdr));

	return ntohs(iph->tot_len);
}

static void fix_iptcp_sum(struct iphdr *iph, struct tcphdr *tcph)
{
	unsigned int sum;

	iph->check = 0;
	sum = half_cksum ((void *) iph, sizeof (struct iphdr));
	iph->check = compute_sum(sum);
	tcph->check = 0;
	sum = half_cksum ((void *) &iph->saddr, 8);
	sum += htons (IPPROTO_TCP + sizeof(struct tcphdr)); 
	sum += half_cksum ((void *) tcph, sizeof (struct tcphdr));
	tcph->check = compute_sum (sum);
}

/*
 * sniffjoke_tell, what's sniffjoke tell to do at our ulog engine ? 
 * tell him which ttl should be used!
 *
 * I've make a struct because could be some other information
 * could be returned, in a different future
 */
struct sniffjoke_tell
{
	unsigned char ttl;
#define INFAMOUS_HACK	666
	unsigned int hack_n;
};

/*
 * this struct is used for track tcp session inside sniffjoke, I track
 * tcp session for find the useful TTL for expiring packets, for
 * speak a better english, for discern secrets and be a were-packets
 */
struct sniffjoke_track
{
	unsigned int daddr;
	unsigned short sport, dport;
	unsigned char last_ttl;
	unsigned char last_hack;
	unsigned char exp_rst;
	unsigned int packet_number;
#define ACTIVE	1
#define EXPIRED	2
#define CLOSED	3
	unsigned int status;
};
#define SESSIONTRACK	80
static struct sniffjoke_track alist[SESSIONTRACK];

#define RESET_ON_EXPIRE	3

/* get_track is used for get tcp session insede track list */
static struct sniffjoke_track *get_track_tcpsex(struct iphdr *iph, struct tcphdr *tcph, int *first_free)
{
	int i;
	int ffset =0, first_closed =0, fclose =0;

	sniffjokelog("get_track_tcpsex: looking for %s daddr\n", inet_ntoa( *((struct in_addr *)&iph->daddr) ));
	for(i =0; i < SESSIONTRACK; i++)
	{
		/*
		 * if, you are lookig for the first free giving a pointer,
		 * if, first free is not already set
		 * if, alist[i].status is 0, 
		 *
		 * wow: you've found the first element free in the array :P
		 */
		if(first_free != NULL && (!ffset) && (!alist[i].status) )
		{
			*first_free =i;
			ffset++;
			sniffjokelog("get_track_tcpsex: *first_free set %d, first closed %d\n", i, first_closed);
		}

		if(!fclose && alist[i].status ==CLOSED)
		{
			first_closed =i;
		}

		if(alist[i].daddr == iph->daddr &&
		   alist[i].sport == tcph->source &&
		   alist[i].dport == tcph->dest )
		{
			/* 
			 * we have found it! 
			 * ip->id is checked before the decrement
			 */
			sniffjokelog("get_track_tcpsex: found session tracked (%u %u %u) %d\n", iph->daddr, tcph->source, tcph->dest, i);
			return &alist[i];
		}
	}

	if(first_free !=NULL && !ffset && first_closed)
	{
		sniffjokelog("get_track_tcpsex: first_free 0, first_closed %d, return %d as empty\n", first_closed, first_closed);
		*first_free =first_closed;
	}

	if(first_free !=NULL)
		memset((void *)&alist[*first_free], 0x00, sizeof(struct sniffjoke_track));

	return NULL;
}

/*
 * this function is used for understand the packets wich state of live has 
 * traveled, the young packets could be ACTIVE, and be decremented for 
 * understand him ttl limits.
 *
 * half-age packets became EXPIRED, here him flowrish live give the best fruits,
 * having exactls the ttl distance with the remote op they could be send with 
 * our most powerful anti sniff tech
 *
 * old packets is know as CLOSED, they stay for teach to youger packets the 
 * role of decrement, the mean of Time To Live and other network philosophy. 
 */
#define DONTSEND	777
static int get_conntrack_or_create_new_entry(struct iphdr *iph, struct sniffjoke_tell *verdict)
{
	struct sniffjoke_track *tracked =NULL;
	struct icmphdr *icmph;
	struct tcphdr *tcph;
	int ffree =0;

	/* sanity check, I need almost some byte for work :P */
	if( ntohs(iph->tot_len) < sizeof(struct iphdr) + sizeof(struct icmphdr))
	{
		sniffjokelog("get_conntrack_or_create_new_entry: return DONTSEND for size limit\n");
		return DONTSEND;
	}

	if(iph->protocol == IPPROTO_TCP)
	{
		tcph =(struct tcphdr *)((unsigned char *)iph + (iph->ihl * 4));
		/* 
		 * TCP is the unique condition for stay in
		 * get_conntrack_or_create_new_entry
		 */
	}
	else if(iph->protocol == IPPROTO_ICMP)
	{
		icmph =(struct icmphdr *)((unsigned char *)iph + (iph->ihl * 4));
		struct iphdr *badiph;
		struct tcphdr *badtcph;

		/* 
		 * we appreciate incoming ICMP but have not time to talk
		 * with non time-exceeded packets, we have to ovecome 
		 * sniffers, not be friendly
		 */
		if(icmph->type != ICMP_TIME_EXCEEDED)
		{
			sniffjokelog("get_conntrack_or_create_new_entry: ICMP bad found (type %d)\n", icmph->type);
			return DONTSEND;
		}

		sniffjokelog("get_conntrack_or_create_new_entry: ICMP TIME EXCEEDED!\n", icmph->type);
		/*
		 * else, icmp time exceeded has icmp data inside, we 
		 * want analyze it
		 */
		badiph =(struct iphdr *)(((unsigned char *)iph + (iph->ihl * 4) + sizeof(struct icmphdr)));
		badtcph =(struct tcphdr *)((unsigned char *)badiph + (badiph->ihl *4));
		/*
		 * tcphdr contains the headers of expired ttl packet,
		 * take a look if was "our" packets
		 */
		tracked =get_track_tcpsex(badiph, badtcph, NULL);

		/* maybe, sadly, icmp data is not our packet */
		if(tracked == NULL)
		{
			sniffjokelog("get_conntrack_or_create_new_entry: not our icmp time exceeded, sadly :(\n");
			return DONTSEND;
		}

		/* else, our ACTIVE state change, is EXPIRED and you know it! */
		tracked->status =EXPIRED;
		verdict->ttl =tracked->last_ttl;
		verdict->hack_n =INFAMOUS_HACK;
		sniffjokelog("get_conntrack_or_create_new_entry: EXPIRED status, verdict ttl %d\n", verdict->ttl);
		return DONTSEND;
	}
	else
	{
		sniffjokelog("get_conntrack_or_create_new_entry: nor TCP/ICMP, return DONTSEND\n");
		return DONTSEND;
	}

	/* we are tcp ? so, get tcp session */
	tracked =get_track_tcpsex(iph, tcph, &ffree);

	/* are you newborn ? welcome ! */
	if(tracked == NULL)
	{
		sniffjokelog("get_conntrack_or_create_new_entry: new TCP session found! position %d (%u %u %u) verdict ttl %d\n", 
			ffree, iph->daddr, tcph->source, tcph->dest, iph->ttl - 1
		);
		alist[ffree].status =ACTIVE;
		alist[ffree].daddr =iph->daddr;
		alist[ffree].sport =tcph->source;
		alist[ffree].dport =tcph->dest;
		alist[ffree].last_ttl =(iph->ttl -1);
		alist[ffree].packet_number =0;

		/* fix verdict struct and affide our first hack */
		verdict->ttl =(iph->ttl -1);
		verdict->hack_n =1; 

		return ACTIVE;
	}

	/* else, is a know connection ... */
	if(!tracked->last_ttl)
		tracked->status =CLOSED;

	/* LAN or GW ? */
	if(tracked->last_ttl == 2)
	{
		sniffjokelog("get_conntrack_or_create_new_entry found ttl 2 and desume network incesto (DONTSEND)\n");
		tracked->status =CLOSED;
	}

	if(tracked->status == EXPIRED)
	{
		if(tracked->exp_rst == RESET_ON_EXPIRE)
		{
			sniffjokelog("get_conntrack_or_create_new_entry EXPIRED limit reached, session became CLOSEDd\n");
			tracked->status = CLOSED;
		}
		else
		{
			tracked->exp_rst++;
			sniffjokelog("get_conntrack_or_create_new_entry found EXPIRED session, send %d RST on %d\n", tracked->exp_rst, RESET_ON_EXPIRE);
		}
	}

	if(tracked->status == CLOSED)
	{
		sniffjokelog("get_conntrack_or_create_new_entery found elder tcp, status CLOSED, return DONTSEND\n");
		return DONTSEND;
	}

	/* else, decrement last ttl until expire */
	tracked->last_ttl--;
	tracked->packet_number++;

	/* each TWO packet we put our TCP hack, to avoid CWND resize */
	if(tracked->packet_number % 2)
	{
		/* 
		 * simple we try one hack only for type, is sniffer are
		 * vulnerable, one packet is enought
		 */
		if(tracked->last_hack == SNIFFJOKE_HACK_NUMBER)
			tracked->last_hack = 0;

		verdict->hack_n = ++(tracked->last_hack);
	}
	else
	{
		/* one packet in two don't has ttl dec and is not hacked */
		tracked->last_ttl++;
		return DONTSEND;
	}

#if 0
	/* each packet has him hack */
	if(tracked->last_hack == SNIFFJOKE_HACK_NUMBER)
		tracked->last_hack = 0;

	verdict->hack_n = ++(tracked->last_hack);
#endif

	verdict->ttl =tracked->last_ttl;
	sniffjokelog("get_conntrack_or_create_new_entery found session and decrement ttl for find him life: id %u ttl %d, answer hack %d\n",
		iph->id, verdict->ttl, verdict->hack_n
	);
	return ACTIVE;
}

static int sniffjoke_hack(ulog_iret_t *ret)
{
	unsigned int pktlen =GET_VALUE(1).ui32, sum, whats;
	struct iphdr *iph = (struct iphdr *)GET_VALUE(0).ptr;
	struct tcphdr *tcph;
	struct sockaddr_in sa;
	struct sniffjoke_tell verdict;

	/* 
	 * packets ? wow, 
	 * should be I must hack it, 
	 * should be I've finished with this session
	 * should be I've already sent this packet, and looping with netfilter
	 *
	 * at vodka's house I've find solution for:
	 *
	 * 1) avoid netfilter/userspace loop
	 * 2) implement ttl hack with few difficult
	 * 3) save bandwidth 
	 * 4) user hacks packet as probe for the uniqe sure tech
	 *
	 * verdict contains ttl to use, 
	 */

	whats =get_conntrack_or_create_new_entry(iph, &verdict);

	/*
	 * happens DONTSEND when icmp/udp packet is read, packets 
	 * is already sent or session is already hacked
	 */
	if(whats ==DONTSEND)
		return 0;

	/*
	 * ok, only tcp survive here
	 */
	tcph =(struct tcphdr *)((unsigned char *)iph + (iph->ihl * 4));

	if(tcph->syn)
	{
		sniffjokelog("sniffjoke_hack: tcph->syn, return without hack, I prefer keep syn easy\n");
		return 0;
	}

	/* any cases want to use ttl verdicted */
	iph->ttl =verdict.ttl;

	/* EXPIRED ? so we could apply or preferred hack */
	if(whats == EXPIRED)
	{
		sniffjokelog("sniffjoke_hack: EXPIRED, try to build our reset with ttl %d\n", iph->ttl);

		tcph->res2 = tcph->res1 = tcph->fin = tcph->syn = tcph->psh = tcph->urg = 0;
		tcph->rst = tcph->ack = 1;
		tcph->ack_seq =tcph->seq;
		tcph->urg_ptr = 0;

		/* iph->ttl is fixed jet */

		/* truncate tcp data & fix checksum */
		pktlen =truncate_tcp_data(tcph, iph);
		fix_iptcp_sum(iph, tcph);

		/* jump for send our friendly reset :) */
		goto send_packet;
	}

	/* 
	 * because duplicate packet with bad checksum cause CWND decrementation,
	 * only one packets in two is apply with hacks
	 */
	if(!verdict.hack_n)
		return 0;

	/* else, ACTIVE, try other hacks */
	switch(verdict.hack_n)
	{
		case 1: // FAKE_CKSUM_RST:
#if 1
			sniffjokelog("sniffjoke_hack: invalid (tcp) checksum with RST inject - [remote CWND could be resize]\n");
			tcph->rst =1;
			tcph->psh =0;
			tcph->ack =0;

			tcph->check ^= 0xd34d;

			/* for RST packet, truncation of data payload */
			pktlen =truncate_tcp_data(tcph, iph);

			/* for better appearence, ip cksum is valid */
			iph->check = 0;
			sum = half_cksum ((void *) iph, sizeof (struct iphdr));
			iph->check = compute_sum(sum);
#else
			iph =NULL;
#endif
			break;
		case 2: // FAKE_CKSUM_FIN:
#if 1
			sniffjokelog("sniffjoke_hack: invalid tcp checksum with FIN injection\n");
			tcph->fin =1;
			tcph->psh =0;
			tcph->check ^= 0xd34d;

			/* for better appearence, ip cksum is valid */
			iph->check = 0;
			sum = half_cksum ((void *) iph, sizeof (struct iphdr));
			iph->check = compute_sum (sum);
#else
			iph =NULL;
#endif
			break;
		case 3: // FAKE_CKSUM_SEQ:
#if 1 
			sniffjokelog("sniffjoke_hack: invalid tcp checksum with bad sequence number\n");
			tcph->seq += tcph->window + 4096;
			/* for better appearence, ip cksum is valid */
			iph->check = 0;
			sum = half_cksum ((void *) iph, sizeof (struct iphdr));
			iph->check = compute_sum(sum);
#else
			iph = NULL;
#endif
			break;
		case 4: // FAKE_SEQEN_RST: 
#if 1
			sniffjokelog("sniffjoke_hack: valid reset with fake sequence inject\n");
			tcph->seq += tcph->window + 4096;
			tcph->psh =0;
			tcph->rst =1;
			tcph->ack =1;
			tcph->ack_seq =tcph->seq;
			/* for RST packet, truncation of data payload */

			pktlen =truncate_tcp_data(tcph, iph);
			/* for better appearence, ip & tcp cksum is valid */
			fix_iptcp_sum(iph, tcph);
#else
			iph = NULL;
#endif
			break;
		case 5: // FAKE_POSTSY_SEQ:
#if 1
			sniffjokelog("sniffjoke_hack: repeat valid syn in established connection\n");
			tcph->syn =1;
			tcph->psh =0;
			tcph->ack =0;
			tcph->ack_seq =0;
			tcph->seq = (tcph->seq + (tcph->window) + 1024);

			pktlen =truncate_tcp_data(tcph, iph);
			/* for better appearence, ip & tcp cksum is valid */
			fix_iptcp_sum(iph, tcph);
#else
			iph =NULL;
#endif
			break;
		case 6: // EVIL_TCPTRUNC:
#if 1
			sniffjokelog("sniffjoke_hack: evil tcp truncation\n");
			iph->tot_len =htons( (iph->ihl * 4) + (tcph->doff * 4) - 4);
			pktlen = ntohs(iph->tot_len);

			/* for better appearence, ip & tcp cksum is valid */
			fix_iptcp_sum(iph, tcph);
#else
			iph =NULL;
#endif
			break;
		case 7: // SNIFFJOKE_GAME:
#if 1
			sniffjokelog("sniffjoke_hack: sniffjoke randomize reserved flag res2 (2 bits) res1 (4 bits)\n");
			tcph->res2 = (rand() % 4);
			tcph->res1 = (rand() % 16);
			tcph->rst =tcph->syn =tcph->fin =0;
			tcph->seq += (tcph->window + 1024);
			tcph->ack_seq += (tcph->window + 1024);

			/* for better appearence, ip & tcp cksum is valid */
			fix_iptcp_sum(iph, tcph);
#else
			iph =NULL;
#endif
			break;
		case 8: // SNIFFJOKE_GAME_2:
#if 1
			sniffjokelog("sniffjoke_hack: sniffjoke fool/doff incrementation\n");
			tcph->doff++;

			/* for better appearence, ip & tcp cksum is valid */
			fix_iptcp_sum(iph, tcph);
#else
			iph = NULL;
#endif
			break;
		default:
			sniffjokelog("sniffjoke_hack: world hasn't certain anymore...(hack_n %d ?)\n", verdict.hack_n);
	}

send_packet:
	/* iph nullity happens in developer's debug for tcp hack */
	if(iph !=NULL)
	{
		sa.sin_addr.s_addr = iph->daddr;
		sa.sin_port = tcph->dest;
		sa.sin_family = PF_INET;

		sniffjokelog("sendto: injecting %d byte hack %d\n\n", pktlen, verdict.hack_n);

		if((sendto (sockfd, (void *)iph, pktlen, 0, (struct sockaddr *)&sa, sizeof (sa))) == -1)
		{
			sniffjokelog("unable to send %d bytes over socket raw (%s) hack type %d\n\n", pktlen, strerror(errno), verdict.hack_n);
			return 0;
		}
	}

	return 0;
}

static void sniffjoke_signal_handler(int signal)
{
	ulogd_log(ULOGD_NOTICE, "sniffjoke_signal_handler, signal %d, what I should do ?\n", signal);
}

static int sniffjoke_init(void)
{
	struct passwd *dropuser;
	int hdrincl =1, uid;
	const char delimit=',';
	char *opt_port;

	if ((sockfd = socket (PF_INET, SOCK_RAW, IPPROTO_TCP)) == -1)
	{
		ulogd_log(ULOGD_ERROR, "unable to open socket raw: %s\n", 
			strerror(errno)
		);
		return 1;
	}
	setsockopt(sockfd, SOL_IP, IP_HDRINCL, &hdrincl, sizeof (int));

	if((log =fopen(fileopt_ce.u.string, "a+")) ==NULL)
	{
		ulogd_log(ULOGD_ERROR, 
		"unable to open logfile %s: %s, using ulogd for sniffjoke\n",
		fileopt_ce.u.string, strerror(errno)
			);
		return 1;
	}
	else
	{
		ulogd_log(ULOGD_NOTICE,
		"starting SNIFFJOKE log to %s\n", fileopt_ce.u.string
			);
	}

	if(!strcmp(useropt_ce.u.string, SNIFFJOKE_DEFAULT_USER))
	{
		ulogd_log(ULOGD_ERROR,
		"sniffjoke require a valid user where apply iptables MATCH extension\n"
			);
		return 1;
	}

	sniffjokelog("SNIFFJOKE started with options: logfile [%s] user [%s] tcp ports [%s] ",
		fileopt_ce.u.string, useropt_ce.u.string, portopt_ce.u.string
	);

	srand(time(NULL));

	if((uid =getuid()) != 0)
	{
		ulogd_log(ULOGD_ERROR, "sniffjoke require root permission in setup\n");
		return 1;
	}

	if((dropuser =getpwnam(useropt_ce.u.string)) ==NULL)
	{
		ulogd_log(ULOGD_ERROR, "unable to find user %s\n", useropt_ce.u.string);
		return 1;
	}

	if(!strlen(portopt_ce.u.string))
	{
		ulogd_log(ULOGD_ERROR, "unable to set TCP port without port option\n");
		return 1;
	}

	opt_port =strtok(portopt_ce.u.string, &delimit);

	while(1)
	{
		if(opt_port !=NULL)
		{
			rules++;

			ipt_put_tcp =realloc( (void *)ipt_put_tcp, rules * sizeof(unsigned int));
			ipt_put_tcp[rules -1] =malloc(CMDLINE);
			ipt_remove_tcp =realloc( (void *)ipt_remove_tcp, rules * sizeof(unsigned int));
			ipt_remove_tcp[rules -1] =malloc(CMDLINE);
			
			snprintf(ipt_put_tcp[rules -1], CMDLINE, 
				"iptables -A OUTPUT -p tcp --dport %s -m owner --uid-owner %d -j ULOG",
				opt_port, dropuser->pw_uid
			);
			snprintf(ipt_remove_tcp[rules -1], CMDLINE, 
				"iptables -D OUTPUT -p tcp --dport %s -m owner --uid-owner %d -j ULOG",
				opt_port, dropuser->pw_uid
			);

			sniffjokelog("sniffjoke add ipt rule for hack tcp outgoing packets [%s]\n", ipt_put_tcp[rules -1]);
			system(ipt_put_tcp[rules -1]);
		}
		else
			break;

		opt_port =strtok(NULL, &delimit);
	}

	snprintf(ipt_put_icmp, CMDLINE, "iptables -A INPUT -p icmp --icmp-type time-exceeded -j ULOG");
	snprintf(ipt_remove_icmp, CMDLINE, "iptables -D INPUT -p icmp --icmp-type time-exceeded -j ULOG");
	sniffjokelog("sniffjoke add ipt rule for read about ttl expired packets [%s]\n", ipt_put_icmp);
	system(ipt_put_icmp);

	return 0;
}

static void sniffjoke_fini(void)
{
	int i;

	ulogd_log(ULOGD_NOTICE, "strippin iptables icmp [%s]\n", ipt_remove_icmp);
	system(ipt_remove_icmp);

	for(i =0; i < rules; i++)
	{
		ulogd_log(ULOGD_NOTICE, "strippin iptables tcp [%s]\n", ipt_remove_tcp[i]);
		system(ipt_remove_tcp[i]);
	}
}

static ulog_output_t sniffjoke_op = {
        .name = "sniffjoke",
	.init = &sniffjoke_init,
	.fini = &sniffjoke_fini,
	.output = &sniffjoke_hack,
	.signal = &sniffjoke_signal_handler,
};

/* get all key id's for the keys we are intrested in */
static int get_ids(void)
{
	int i;
	struct intr_id *cur_id;

	for (i = 0; i < INTR_IDS; i++) 
	{
		cur_id = &intr_ids[i]; 
		cur_id->id = keyh_getid(cur_id->name);
		if (!cur_id->id) 
		{
			ulogd_log(ULOGD_ERROR, "Cannot resolve keyhash id for %s\n", cur_id->name);
			return 1;
		}
	}
	return 0;
}

void _init(void)
{
	if (get_ids()) 
	{
		ulogd_log(ULOGD_ERROR, "can't resolve all keyhash id's\n");
	}

	config_parse_file("SNIFFJOKE", &fileopt_ce);
	config_parse_file("SNIFFJOKE", &portopt_ce);
	config_parse_file("SNIFFJOKE", &useropt_ce);

	register_output(&sniffjoke_op);
}
